/*
 * iap_fup.c - update firmware using IAP functions
 *   by uratan! 2008.11.4
 */
#include "common.h"
#include "LPC214x.h"

/*
 * my local debug flag
 */
#define myUDEBUG  (0 && gUDEBUG)
#define RPRINTF    if(myUDEBUG)rprintf
	/*
	 * as the codes of rprintf() function is already erased,
	 * !! you can NOT debug iap_update_on_ram() with rprintf() !!
	 */

/*
 * FLASH SECTOR MAP                           LPC2141(32kB)
 *                                            | LPC2142(64kB)
 *    Sector Number                           | | LPC2144(128kB)
 *     |   Sector Size                        | | | LPC2146(256kB)
 *     |    |      Address Range              | | | | LPC2148(512kB)
 *     |    |       |                         | | | | |
 *  +----+------+---------------------------+-----------+
 *  |  0 |  4kB | 0x0000_0000 - 0x0000_0FFF | + + + + + |
 *  |  1 |  4kB | 0x0000_1000 - 0x0000_1FFF | + + + + + |
 *  |  2 |  4kB | 0x0000_2000 - 0x0000_2FFF | + + + + + |
 *  |  3 |  4kB | 0x0000_3000 - 0x0000_3FFF | + + + + + |
 *  |  4 |  4kB | 0x0000_4000 - 0x0000_4FFF | + + + + + |
 *  |  5 |  4kB | 0x0000_5000 - 0x0000_5FFF | + + + + + |
 *  |  6 |  4kB | 0x0000_6000 - 0x0000_6FFF | + + + + + |
 *  |  7 |  4kB | 0x0000_7000 - 0x0000_7FFF | + + + + + |
 *  |  8 | 32kB | 0x0000_8000 - 0x0000_FFFF |   + + + + |
 *  |  9 | 32kB | 0x0001_0000 - 0x0001_7FFF |     + + + |
 *  | 10 | 32kB | 0x0001_8000 - 0x0001_FFFF |     + + + |
 *  | 11 | 32kB | 0x0002_0000 - 0x0002_7FFF |       + + |
 *  | 12 | 32kB | 0x0002_8000 - 0x0002_FFFF |       + + |
 *  | 13 | 32kB | 0x0003_0000 - 0x0003_7FFF |       + + |
 *  | 14 | 32kB | 0x0003_8000 - 0x0003_FFFF |       + + |
 *  | 15 | 32kB | 0x0004_0000 - 0x0004_7FFF |         + |
 *  | 16 | 32kB | 0x0004_8000 - 0x0004_FFFF |         + |
 *  | 17 | 32kB | 0x0005_0000 - 0x0005_7FFF |         + |
 *  | 18 | 32kB | 0x0005_8000 - 0x0005_FFFF |         + |
 *  | 19 | 32kB | 0x0006_0000 - 0x0006_7FFF |         + |
 *  | 20 | 32kB | 0x0006_8000 - 0x0006_FFFF |         + |
 *  | 21 | 32kB | 0x0007_0000 - 0x0007_7FFF |         + |
 *  | 22 |  4kB | 0x0007_8000 - 0x0007_8FFF |         + |
 *  | 23 |  4kB | 0x0007_9000 - 0x0007_9FFF |         + |
 *  | 24 |  4kB | 0x0007_A000 - 0x0007_AFFF |         + |
 *  | 25 |  4kB | 0x0007_B000 - 0x0007_BFFF |         + |
 *  | 26 |  4kB | 0x0007_C000 - 0x0007_CFFF |         + |
 *  +----+------+---------------------------+-----------+
 */
/*
 * divide FLASH to 2 areas
 *   one for master
 *   the other for temporary place to update firmware
 */
#define FL_MASTER_TOP_ADDR    0x0000
#define FL_MASTER_END_ADDR    0x3fff
#define FL_WORKING_TOP_ADDR   0x4000	/* until 0x7fff */

/*
 * my using-IAP definitions
 */
#define BYTES_per_PAGE  256	/* [bytes] */

/*
 * IAP entry (IAP is thumb code)
 */
#define IAP_LOCATION  0x7ffffff1

/*
 * IAP commands
 */
#define IAPCM_PREPARE_WRITE_OPERATION   50
#define IAPCM_COPY_RAM_TO_FLASH         51
#define IAPCM_ERASE_SECTORS             52
#define IAPCM_BLANK_CHECK_SECTORS       53
#define IAPCM_READ_PART_ID              54
#define IAPCM_READ_BOOT_CODE_VERSION    55
#define IAPCM_COMPARE                   56
#define IAPCM_REINVOKE_ISP              57

/*
 * IAP status
 */
#define IAPST_CMD_SUCCESS           0	/* Command is executed successfully */
#define IAPST_INVALID_COMMAND       1	/* Invalid command */
#define IAPST_SRC_ADDR_ERROR        2	/* not on a word boundary */
#define IAPST_DST_ADDR_ERROR        3	/* not on a correct boundary */
#define IAPST_SRC_ADDR_NOT_MAPPED   4	/* not mapped in the memory map */
#define IAPST_DST_ADDR_NOT_MAPPED   5	/* not mapped in the memory map */
#define IAPST_COUNT_ERROR           6	/* Byte count is wrong */
#define IAPST_INVALID_SECTOR        7	/* Sector number is invalid */
#define IAPST_SECTOR_NOT_BLANK      8	/* Sector is not blank */
#define IAPST_SECTOR_NOT_PREPARED   9	/* write preparation was not done */
#define IAPST_COMPARE_ERROR        10	/* data is not same */
#define IAPST_BUSY                 11	/* programming hardware is busy */

/*
 * for IAP function call
 */
typedef void (*T_IAP)(u_long command[], u_long result[]);

/*
 * functions
 */
void iap_fup_init(void);
int fup_erase(void);
int fup_write(u_char *buff);
void fup_update_master(void);
void iap_update_on_ram(void)              __attribute__ ((section (".fup")));
static int iap_erase_all(u_long top_addr) __attribute__ ((section (".fup")));
static int iap_write_page(u_long addr)    __attribute__ ((section (".fup")));
/*static*/ void iap_reboot(void)          __attribute__ ((section (".fup")));
static u_long addr_to_sect(u_long addr)   __attribute__ ((section (".fup")));

/*
 * variables
 */
static u_long fup_write_addr;
static union {
	u_long word[BYTES_per_PAGE/sizeof(u_long)];
	u_char byte[BYTES_per_PAGE];
} iap_buff;

static u_long command[5];
static u_long result[2];
static const T_IAP iap_entry = (T_IAP)IAP_LOCATION;

/*
 *
 */
void iap_fup_init(void)
{
//	iap_entry = (T_IAP)IAP_LOCATION;
	WDMOD = 0x00;
		/*
		 * clear watch-dog timr-out flag
		 *   never enter ISP mode if this flag is remained set
		 */
}

/*
 * erase working area
 */
int fup_erase(void)
{
	int stat;
	u_long imf;

	/*
	 * rewind address counter for write
	 */
	fup_write_addr = FL_WORKING_TOP_ADDR;

	/*
	 * erase working area
	 */
	imf = IntDisable();
	stat = iap_erase_all(FL_WORKING_TOP_ADDR);
	IntRestore(imf);

	if(stat != IAPST_CMD_SUCCESS)
		return(-1);

	return(0);
}

/*
 * write data onto working area
 */
int fup_write(u_char *buff)
{
	u_char *dst;
	int i, stat;
	u_long imf;

	/*
	 * copy data to my buffer
	 */
	dst = &iap_buff.byte[fup_write_addr % BYTES_per_PAGE];
	for(i=0; i<8; i++) {
		*dst++ = *buff++;
	}

	/*
	 * advance address counter
	 */
	fup_write_addr += 8;

	/*
	 * write working area if all the page is filled
	 */
	if((fup_write_addr % BYTES_per_PAGE) == 0x00) {
		imf = IntDisable();
		stat = iap_write_page(fup_write_addr - BYTES_per_PAGE);
		IntRestore(imf);

		if(stat != IAPST_CMD_SUCCESS)
			return(-1);
	}

	return(0);
}

/*
 * copy working area onto master, then reboot
 */
void fup_update_master(void)
{
	/*
	 * I expect:
	 *   1) all functions needed to this operation have
	 *      "__attribute__ ((section (".fup*")))"
	 *   2) functions are chained by relative-branch.
	 *   3) code sizes in total is less than exec_buffer size.
	 *
	 * see 'Startup.S' about copy_n_exec_on_ram();
	 */
	extern void copy_n_exec_on_ram(void *x);

	IntDisable();

	/*
	 * copy code on RAM, execute them from RAM
	 */
	copy_n_exec_on_ram(iap_update_on_ram);
		/* never return */
}

/*
 * copy working area onto master, then reboot
 * !! this function shall be run from RAM, not from FLASH
 */
void iap_update_on_ram(void)
{
	u_long stat;
	u_long dst_addr, *src_ptr, i;

	/*
	 * erase master area
	 */
	stat = iap_erase_all(FL_MASTER_TOP_ADDR);

	/*
	 * copy all working area to master
	 */
	dst_addr = FL_MASTER_TOP_ADDR;
	src_ptr = (u_long *)FL_WORKING_TOP_ADDR;

	while(stat == IAPST_CMD_SUCCESS) {
		/*
		 * load data into my buffer
		 */
		for(i=0; i<(BYTES_per_PAGE/sizeof(u_long)); i++) {
			iap_buff.word[i] = *src_ptr++;
		}

		/*
		 * do write
		 */
		stat = iap_write_page(dst_addr);

		/*
		 * advance pointer until end
		 */
		dst_addr += BYTES_per_PAGE;
		if(dst_addr > FL_MASTER_END_ADDR)
			break;
	}

	/*
	 * reset myself if success to 
	 */
	if(stat == IAPST_CMD_SUCCESS) {
		iap_reboot();
			/* never return */
	} else {
		/*  freeze */
		while(1)
			;
	}

	return;
}

/*
 * erase multiple sectors of FLASH
 */
static int iap_erase_all(u_long top_addr)
{
	u_long end_addr;

	/*
	 * make end_addr by relative
	 */
	end_addr = top_addr + (FL_MASTER_END_ADDR - FL_MASTER_TOP_ADDR);

	/*
	 * do preparation
	 */
	command[0] = IAPCM_PREPARE_WRITE_OPERATION;
	command[1] = addr_to_sect(top_addr);
	command[2] = addr_to_sect(end_addr);
	iap_entry(command, result);
RPRINTF("PWO: %d\r\n", result[0]);

	/*
	 * abandon operation if something is wrong
	 */
	if(result[0] != IAPST_CMD_SUCCESS)
		return(result[0]);

	/*
	 * do erase
	 */
	command[0] = IAPCM_ERASE_SECTORS;
	command[1] = addr_to_sect(top_addr);
	command[2] = addr_to_sect(end_addr);
	command[3] = Fcclk / 1000;
	iap_entry(command, result);
RPRINTF("PWO: %d\r\n", result[0]);

	return(result[0]);
}

/*
 * write iap_buff onto FLASH
 */
static int iap_write_page(u_long addr)
{
	/*
	 * do preparation
	 */
	command[0] = IAPCM_PREPARE_WRITE_OPERATION;
	command[1] = addr_to_sect(addr);
	command[2] = command[1];
	iap_entry(command, result);
RPRINTF("PWO: %d\r\n", result[0]);

	/*
	 * abandon operation if something is wrong
	 */
	if(result[0] != IAPST_CMD_SUCCESS)
		return(result[0]);

	/*
	 * do write
	 */
	command[0] = IAPCM_COPY_RAM_TO_FLASH;
	command[1] = addr;
	command[2] = (u_long)iap_buff.word;
	command[3] = BYTES_per_PAGE;
	command[4] = Fcclk / 1000;
	iap_entry(command, result);
RPRINTF("PWO: %d\r\n", result[0]);

	return(result[0]);
}

/*
 * reset myself
 */
void iap_reboot(void)
{
	/*
	 * use WDT to make RESET
	 */
	WDTC = 0x00ff;		/* minimum timeout */
	WDMOD = 0x03;		/* enable to reset cpu */
	WDFEED = 0xaa;		/* once feed to run WDT */
	WDFEED = 0x55;

	/*
	 * wait for RESET
	 */
	while(1)
		;

	/* never come here */
}

/*
 * convert FLASH address to sector number
 */
static u_long addr_to_sect(u_long addr)
{
	u_long sect_num;

	if(addr >= 0x00078000) {
		sect_num = 22 + (addr - 0x00078000) / 0x1000;
	} else if(addr >= 0x00008000) {
		sect_num =  8 + (addr - 0x00008000) / 0x8000;
	} else {
		sect_num =  0 + (addr - 0x00000000) / 0x1000;
	}

	return(sect_num);
}
