/*
 * main.c : main C entry
 *   by uratan! 2008.9.8
 */
	/* 
	 * LPC2148 ERRATA SHEET (v1.8/Jul 7, 2008) has:
	 *   Core.1 : Incorrect update of the Abort link register
	 *   MAM.1  : Incorrect read of data from SRAM
	 *   MAM.2  : Code execution failure can occur with MAM Mode 2
	 *   SPI.1  : Incorrect shifting of data in slave mode at lower freq
	 *   SSP.1  : Initial data bits/clocks corrupted in SSP transmission
	 *   Timer.1: Timer Counter reset occurs on incorrect edge in counter mode
	 *   DC/DC.1: DC/DC Converter start-up issue
	 *   USB.1  : USB interface does not function if P0.23 is held low
	 *   WDT.1  : Accessig non-Watchdog APB registers in the middle of ...
	 *    etc.  :  ...
	 *
	 *    +---------------+
	 *    |    N X P      |
	 *    | LPC2148FBD64  |
	 *    | SF7521.1  06  |
	 *    | ZPG0729A      | (my LPC2148 may be revision-A)
	 *    |               |
	 *    |o              |
	 *    +---------------+
	 */
#include "common.h"
#include "LPC214x.h"
#include "irq.h"
#if gUDEBUG
#include "dscr.h.S"
#endif /* gUDEBUG */

/*
 * for uart-debug
 */
#define RPRINTF    if(gUDEBUG)rprintf

/*
 * functions I have
 */
int main(void);
void sys_init(void);
//void IRQ_handler(void)   __attribute__ ((interrupt("IRQ")));
void FIQ_handler(void)   __attribute__ ((interrupt("FIQ")));
//void SWI_handler(void)   __attribute__ ((interrupt("SWI")));
void UNDEF_handler(void) __attribute__ ((interrupt("UNDEF")));
void PABT_handler(void)  __attribute__ ((interrupt("ABORT")));
void DABT_handler(void)  __attribute__ ((interrupt("ABORT")));

/*
 * my variables
 */
VOLATILE int GotSUD;	/* SetUp Data */
VOLATILE int GotCOD;	/* Control-OUT Data */
VOLATILE int Sleep;		/* shall Sleep becaue in Suspend mode */

/*
 * externs
 */
/* in iap_fup.c */
extern void iap_fup_init(void);
/* in pdown.c */
extern void pdown_init(void);
/* in usbhw.c */
extern int uhw_Init(void);
extern void uhw_Connect(int conn);
#if gUDEBUG
extern void uhw_SetStallEP(u_char EPaddr);
#endif /* gUDEBUG */
/* in usbcore.c */
extern void ucore_InitialCore(void);
extern void ucore_TryPowerDown(void);
extern void ucore_SetupCommand(void);
extern void ucore_ProceedSetupCommand(void);
/* in periph.c */
extern void TD_Init(void);
extern void TD_Poll(void);
extern int TD_Suspend(void);
extern void TD_Resume(void);

/* in dscr.S */
extern u_long *bcdDevice, *idVendor, *idProduct;

/*
 * main entry
 */
int main(void)
{
	/*
	 * setup misc system configuration
	 */
	sys_init();
	init_VIC();		/* Vectored Interrupt Controller */

#if gUDEBUG
	/*
	 * setup debug UART port
	 */
	cm_init_serial0(57600);
	rprintf(
		"\r\n"
		"/*\r\n"
		" * start ArmUZK (%d+%dMHz)\r\n"
		" *   ver %04x\r\n"
		" * idVendor: 0x%04x  idProduct: 0x%04x\r\n"
		" */\r\n",
		Fcclk/1000000, Fpclk/1000000,
		&bcdDevice,
		&idVendor, &idProduct
	);
//	rprintf("RSRI is %02x\r\n", RSIR);
//	RSIR = 0x0f;
//	rprintf("RSRI is %02x\r\n", RSIR);
//	rprintf("WDMOD is %02x\r\n", WDMOD);
//	WDMOD = 0x00;		/* never enter ISP if WDTOF is set */
//	rprintf("WDMOD is %02x\r\n", WDMOD);
#endif /* gUDEBUG */

	/*
	 * initial my variables
	 */
	GotSUD = FALSE;
	GotCOD = FALSE;
	Sleep = FALSE;

	/*
	 * initial LED and debug I/O
	 */
	FIO0DIR |= LED_MON;
	MON_LED_ON;

	FIO1DIR |= LED_STAT | DEBUG_PORT | DEBUG_PORT2;
	FIO1SET = LED_STAT;						/* turn off LED */
	FIO1CLR = DEBUG_PORT | DEBUG_PORT2;		/* set low */

	/*
	 * initial peripherals
	 */
	iap_fup_init();
	pdown_init();
	uhw_Init();			/* initial about USB */
	ucore_InitialCore();

	TD_Init();

	/*
	 * here we go
	 */
	IntEnable();			/* permit interrupts */
	uhw_Connect(TRUE);		/* connect USB */

	/*
	 * wait for events for endless
	 */
	while(1) {
		/*
		 * we shall sleep when USB-Suspend detected
		 */
		if(Sleep) {
			if(TD_Suspend()==DONE) {
				Sleep = FALSE;
				ucore_TryPowerDown();
			}
			TD_Resume();
		}

		/*
		 * wait for SETUP data available
		 */
		if(GotSUD) {
			GotSUD = FALSE;			/* Clear SUDAV flag */
SET_DEBUG_PORT;
			ucore_SetupCommand();
CLR_DEBUG_PORT;
		}

		/*
		 * wait for Control-OUT data
		 */
		if(GotCOD) {
			GotCOD = FALSE;
SET_DEBUG_PORT;
			ucore_ProceedSetupCommand();
CLR_DEBUG_PORT;
		}

		/*
		 * call user polling routines
		 */
		TD_Poll();

#if gUDEBUG
{
#define MEMSTART  0x40000000
#define MEMEND    0x40001000
	int i, j, k;

	k = cm_keyin();
	if(k != -1) {
		switch(k) {
		case 'D':
				/* Dump Memories */
				for(i=MEMSTART; i<MEMEND; i+=16) {
					rprintf("%08x:", i);
					for(j=0; j<16; j++) {
						rprintf(" %02x", ((u_char *)i)[j]);
					}
					rprintf("\r\n");
				}
				break;
		case '0':
				uhw_SetStallEP(0x00);	/*EP_CTRL_OUT*/
				uhw_SetStallEP(0x80);	/*EP_CTRL_IN*/
				break;
		case '1':
				uhw_SetStallEP(DSCR_ADDR_EP1i);
				break;
		case '3':
				uhw_SetStallEP(DSCR_ADDR_EP3i);
				break;
		default:
			cm_putchar(k);
		}
	}
}
#endif /* gUDEBUG */
	}

	/* never return from main() */
	while(1)
		;
}


/* =============================================================== */
/*
 * setup core hardware running conditions
 */
void sys_init(void)
{
	/*
	 * confirm reset values
	 */
	MAMCR = 0;		/* disable MAM function */
	MAMTIM = 7;		/* most longer wait */
	VPBDIV = 0;		/* most slow PCLK */

	/*
	 * setup core clock PLL
	 */
	cm_sys_pll_init();
		/*
		 * 011: 60mA
		 *  -PLL: 30mA (-30mA)
		 *  +PER: 64mA ( +4mA)
		 */

	/*
	 * config memory accelerater module
	 */
#if Fcclk < 20000000
	MAMTIM = 1;
#elif Fcclk < 40000000
	MAMTIM = 2;
#else
	MAMTIM = 3;				/* <== */
#endif

	MAMCR = 2;		/* MAM function fully enabled */

	/*
	 * Setting peripheral Clock (PCLK) to System Clock (CCLK)
	 *   0: PCLK = 1 * CCLK / 4
	 *   1: PCLK = 4 * CCLK / 4
	 *   2: PCLK = 2 * CCLK / 4
	 */
#if (Fpclk / (Fcclk / 4)) == 1
	VPBDIV = 0;
#elif (Fpclk / (Fcclk / 4)) == 2
	VPBDIV = 2;
#elif (Fpclk / (Fcclk / 4)) == 4
	VPBDIV = 1;						/* <== */
#else
#error invalid VPBDIV
#endif

	/*
	 * select vector memory
	 */
	MEMMAP = 0x01;		/* User Flash Mode */

	/*
	 * initialize I/O pins
	 */
	SCS |= 0x03;		/* both GPIO are used as Fast I/O */
	PINSEL0 = 0;			/* confirm reset values */
	PINSEL1 = 0;
	FIO0DIR = 0x00000000;
	PINSEL2 = 0;
	FIO1DIR = 0x00000000;

	/*
	 * once turn off all peripherals
	 * you shall turn on them in initialize process of each driver
	 */
	PCONP = 0x00000000;

	return;
}

/* =============================================================== */
/*
 * exception handlers
 */
void FIQ_handler(void)
{
	RPRINTF("!!FIQ\r\n");
	while(1)
		;		/* halt at here */
}

void UNDEF_handler(void)
{
	RPRINTF("!!UNDEF\r\n");
	while(1)
		;		/* halt at here */
}

void PABT_handler(void)
{
	RPRINTF("!!PABT\r\n");
	while(1)
		;		/* halt at here */
}

void DABT_handler(void)
{
	RPRINTF("!!DABT\r\n");
	while(1)
		;		/* halt at here */
}
